"""Classes and helper functions for creating a worker that can run job services in Snowpark Container Services."""

import datetime
import os
import sys
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional
from uuid import UUID

import anyio
import dateutil
import snowflake.connector
import yaml
from prefect_snowflake import SnowflakeCredentials
from pydantic import Field
from slugify import slugify
from snowflake.core import Root
from snowflake.core.exceptions import NotFoundError
from snowflake.core.service import (
    JobService,
    ServiceContainer,
    ServiceResource,
    ServiceSpec,
)

from prefect.client.schemas.objects import FlowRun
from prefect.utilities.asyncutils import run_sync_in_worker_thread
from prefect.utilities.dockerutils import get_prefect_image_name
from prefect.workers.base import (
    BaseJobConfiguration,
    BaseVariables,
    BaseWorker,
    BaseWorkerResult,
)

if TYPE_CHECKING:
    from prefect.server.schemas.core import Flow
    from prefect.server.schemas.responses import DeploymentResponse

SPCS_DEFAULT_CPU_REQUEST = "1"
SPCS_DEFAULT_MEMORY_REQUEST = "1G"
SPCS_DEFAULT_CPU_LIMIT = None
SPCS_DEFAULT_MEMORY_LIMIT = None

DEFAULT_CONTAINER_ENTRYPOINT = "/opt/prefect/entrypoint.sh"


def _get_default_job_manifest_template() -> dict[str, Any]:
    """Returns the default job manifest template used by the SPCS job service."""
    # fmt: off
    return {
        "spec": {
            "containers": [
                {
                    "name": "{{ name }}",
                    "image": "{{ image }}",
                    "command": "{{ entrypoint }}",
                    "args": " {{ command }}",
                    "env": "{{ env }}",
                    "secrets": "{{ secrets }}",
                    "volumeMounts": "{{ volume_mounts }}",
                    "resources": {
                        "requests": {
                            "cpu": "{{ cpu_request }}",
                            "nvidia.com/gpu": "{{ gpu_count }}",
                            "memory": "{{ memory_request }}"
                        },
                        "limits": {
                            "cpu": "{{ cpu_limit }}",
                            "nvidia.com/gpu": "{{ gpu_count }}",
                            "memory": "{{ memory_limit }}"
                        },
                    },
                }
            ],
            "volumes": "{{ volumes }}",
            "logExporters": {
                "eventTableConfig": {
                    "logLevel": "{{ log_level }}"
                }
            },
            "platformMonitor":
            {
                "metricConfig": {
                    "groups": "{{ metrics_groups }}"
                }
            }
        }
    }
    # fmt: on


class SPCSWorkerConfiguration(BaseJobConfiguration):
    """Defines variables which can be overridden in the work pool.

    Also handles validation of the config, whether set in the work pool or in the deployment.
    """

    job_manifest: dict[str, Any] = Field(
        json_schema_extra={"template": _get_default_job_manifest_template()}
    )

    image: str = Field(
        default_factory=get_prefect_image_name,
        description="The image to use for the Prefect container in the task. This value defaults to a Prefect base image matching your local versions.",
        examples=["docker.io/prefecthq/prefect:3-latest"],
    )
    image_registry: str = Field(
        default=None,
        description="The fully qualified name of the Snowflake image registry.",
    )
    entrypoint: Optional[str] = Field(
        default=DEFAULT_CONTAINER_ENTRYPOINT,
        description=(
            "The entrypoint of the container you wish you run."
            "This value defaults to the entrypoint used by Prefect images and should only be changed when using a custom image that is not based on an official Prefect image."
        ),
    )
    snowflake_credentials: SnowflakeCredentials = Field(
        default_factory=SnowflakeCredentials,
        description="Snowflake credentials to use when creating job services.",
    )
    secrets: list[dict[str, str]] = Field(
        default_factory=list,
        description="Snowflake secrets to inject into the container as env variables or files.",
    )
    external_access_integrations: list[str] = Field(
        default_factory=list,
        description="The names of the external access integrations that the service should be created with.",
    )
    compute_pool: str = Field(
        default=None,
        description="The fully-qualified name of the compute pool to run services in.",
        examples=["common.compute.my_compute_pool"],
    )
    cpu_request: str = Field(
        default=SPCS_DEFAULT_CPU_REQUEST,
        description=f"CPU allocation request for the job service. If not provided, a default value of {SPCS_DEFAULT_CPU_REQUEST} will be used.",
    )
    cpu_limit: str = Field(
        default=SPCS_DEFAULT_CPU_LIMIT,
        description="CPU allocation limit for the job service. If not provided, there will be no limit.",
    )
    gpu_count: Optional[int] = Field(
        default=None,
        description="The number of GPUs to use. If not provided, no GPUs will be used.",
    )
    memory_request: str = Field(
        default=SPCS_DEFAULT_MEMORY_REQUEST,
        description=(
            "Memory allocation request for the job service."
            f"If not provided, a default value of {SPCS_DEFAULT_MEMORY_REQUEST} will be used unless present on the task definition."
        ),
    )
    memory_limit: str = Field(
        default=SPCS_DEFAULT_MEMORY_LIMIT,
        description="Memory allocation limit for the job service. If not provided, will be the same as the request.",
    )
    volume_mounts: list[dict[str, str]] = Field(
        default_factory=list,
        description="Defines where volumes appears in the container.",
    )
    volumes: list[dict[str, Any]] = Field(
        default_factory=list,
        description="Shared file system that can be made available in your container.",
    )
    log_level: Literal["INFO", "ERROR", "NONE"] = Field(
        default="INFO",
        description="The level of application logs that Snowflake should collect and export to your event table.",
    )
    metrics_groups: list[str] = Field(
        default_factory=list,
        description="The list of predefined metrics groups that should be logged to the event table.",
    )
    query_warehouse: Optional[str] = Field(
        default=None, description="The query warehouse of the job service."
    )
    service_comment: Optional[str] = Field(
        default=None,
        description="Specify a comment for the job service. Visible in certain Snowflake logs.",
    )
    stream_output: bool = Field(
        default=False,
        description="Direct flow log output back to the worker's console.",
    )
    pool_start_timeout_seconds: int = Field(
        default=600,
        description="The number of seconds to wait for the compute pool to start before considering the run failed.",
    )
    service_start_timeout_seconds: int = Field(
        default=300,
        description="The number of seconds to wait for the job service to start before considering the run failed.",
    )
    service_watch_poll_interval: int = Field(
        default=5,
        description="The number of seconds to wait between Snowflake API calls while monitoring the state of the service.",
    )

    def prepare_for_flow_run(
        self,
        flow_run: "FlowRun",
        deployment: Optional["DeploymentResponse"] = None,
        flow: Optional["Flow"] = None,
    ) -> None:
        """Prepares the job configuration for a flow run."""
        super().prepare_for_flow_run(flow_run, deployment, flow)

        # There's only one container.
        container = self.job_manifest["spec"]["containers"][0]

        # Ensure an image is set.
        container["image"] = self.image

        # Set the container's environment variables by grabbing them from the worker.
        container["env"] = {**self._base_environment(), **self.env}

        # Convert the entrypoint from a string to a list, because that's what SPCS expects
        if self.entrypoint:
            container["command"] = self.entrypoint.split(" ")

        # Convert the command from a string to a list, because that's what SPCS expects
        if self.command:
            container["args"] = self.command.split(" ")

        # Set memory_limit to the same value as memory_request, if not specified.
        if not self.memory_limit:
            container["resources"]["limits"]["memory"] = self.memory_request

        # Remove the platformMonitor section if there are no metrics groups.
        if not self.metrics_groups:
            self.job_manifest["spec"].pop("platformMonitor")

        if self.pool_start_timeout_seconds < 0:
            raise ValueError(
                "pool_start_timeout_seconds must be a non-negative integer."
            )

        if self.service_start_timeout_seconds < 0:
            raise ValueError(
                "service_start_timeout_seconds must be a non-negative integer."
            )


class SPCSServiceTemplateVariables(BaseVariables):
    """Defines variables which can be overridden by deployments.

    Must include all variables in SPCSWorkerConfiguration. All validation should happen in SPCSWorkerConfiguration.
    """

    image: str = Field(
        default_factory=get_prefect_image_name,
        description="The image to use for the Prefect container in the task. This value defaults to a Prefect base image matching your local versions.",
        examples=["docker.io/prefecthq/prefect:3-latest"],
    )
    image_registry: str = Field(
        default=None,
        description="The fully qualified name of the Snowflake image registry.",
    )
    entrypoint: Optional[str] = Field(
        default=DEFAULT_CONTAINER_ENTRYPOINT,
        description=(
            "The entrypoint of the container you wish you run."
            "This value defaults to the entrypoint used by Prefect images and should only be changed when using a custom image that is not based on an official Prefect image."
        ),
    )
    snowflake_credentials: SnowflakeCredentials = Field(
        default_factory=SnowflakeCredentials,
        description="Snowflake credentials to use when creating job services.",
    )
    secrets: list[dict[str, str]] = Field(
        default_factory=list,
        description="Snowflake secrets to inject into the container as env variables or files.",
    )
    external_access_integrations: list[str] = Field(
        default_factory=list,
        description="The names of the external access integrations that the service should be created with.",
    )
    compute_pool: str = Field(
        description="The fully-qualified name of the compute pool to run services in.",
        examples=["common.compute.my_compute_pool"],
    )
    cpu_request: str = Field(
        default=SPCS_DEFAULT_CPU_REQUEST,
        description=f"CPU allocation request for the job service. If not provided, a default value of {SPCS_DEFAULT_CPU_REQUEST} will be used.",
    )
    cpu_limit: str = Field(
        default=SPCS_DEFAULT_CPU_LIMIT,
        description="CPU allocation limit for the job service. If not provided, there will be no limit.",
    )
    gpu_count: Optional[int] = Field(
        default=None,
        description="The number of GPUs to use. If not provided, no GPUs will be used.",
    )
    memory_request: str = Field(
        default=SPCS_DEFAULT_MEMORY_REQUEST,
        description=(
            "Memory allocation request for the job service."
            f"If not provided, a default value of {SPCS_DEFAULT_MEMORY_REQUEST} will be used unless present on the task definition."
        ),
    )
    memory_limit: str = Field(
        default=SPCS_DEFAULT_MEMORY_LIMIT,
        description="Memory allocation limit for the job service. If not provided, will be the same as the request.",
    )
    volume_mounts: list[dict[str, str]] = Field(
        default_factory=list,
        description="Defines where volumes appears in the container.",
    )
    volumes: list[dict[str, Any]] = Field(
        default_factory=list,
        description="Shared file system that can be made available in your container.",
    )
    log_level: Literal["INFO", "ERROR", "NONE"] = Field(
        default="INFO",
        description="The level of application logs that Snowflake should collect and export to your event table.",
    )
    metrics_groups: list[str] = Field(
        default_factory=list,
        description="The list of predefined metrics groups that should be logged to the event table.",
    )
    query_warehouse: Optional[str] = Field(
        default=None, description="The query warehouse of the job service."
    )
    service_comment: Optional[str] = Field(
        default=None,
        description="Specify a comment for the job service. Visible in certain Snowflake logs.",
    )
    stream_output: bool = Field(
        default=False,
        description="Direct flow log output back to the worker's console.",
    )
    pool_start_timeout_seconds: int = Field(
        default=600,
        description="The number of seconds to wait for the compute pool to start before considering the run failed.",
    )
    service_start_timeout_seconds: int = Field(
        default=300,
        description="The number of seconds to wait for the job service to start before considering the run failed.",
    )
    service_watch_poll_interval: int = Field(
        default=5,
        description="The number of seconds to wait between Snowflake API calls while monitoring the state of the service.",
    )


class SPCSWorkerResult(BaseWorkerResult):
    """Result returned by the SPCSWorker."""


class SPCSWorker(BaseWorker):
    """A Prefect worker that runs flows as service jobs in Snowpark Container Services."""

    type: str = "snowpark-container-service"
    job_configuration = SPCSWorkerConfiguration
    job_configuration_variables = SPCSServiceTemplateVariables
    _description = "Execute flow runs within containers on Snowflake's Snowpark Container Services. Requires a Snowflake account."

    async def run(
        self,
        flow_run: FlowRun,
        configuration: SPCSWorkerConfiguration,
        task_status: Optional[anyio.abc.TaskStatus] = None,
    ) -> BaseWorkerResult:
        """Run a flow as a service job in Snowpark Container Services.

        Args:
            flow_run: The flow run to run.
            configuration: The configuration for the flow run.
            task_status: The task status object for the current task. Used
            to provide an identifier that can be used to cancel the task.

        Returns:
            The result of the flow run.

        """
        connection_parameters = self._get_snowflake_connection_parameters(configuration)

        session = snowflake.connector.connect(**connection_parameters)
        root = Root(session)

        # Create the execution environment and start execution
        job_service_name = await run_sync_in_worker_thread(
            self._create_and_start_service,
            root,
            flow_run,
            configuration,
        )

        if task_status:
            # Use a unique ID to mark the run as started.
            # This ID is later used to tear down infrastructure, if the flow run is cancelled.
            task_status.started(job_service_name)

        # Monitor the execution
        job_status = await run_sync_in_worker_thread(
            self._watch_service,
            job_service_name,
            root,
            configuration,
        )

        # We're done in Snowflake, close the connection.
        session.close()

        exit_code = (
            job_status if job_status is not None else -1
        )  # Get the result of the execution for reporting

        return SPCSWorkerResult(
            status_code=exit_code,
            identifier=job_service_name,
        )

    def _create_and_start_service(
        self,
        root: Root,
        flow_run: FlowRun,
        configuration: SPCSWorkerConfiguration,
    ) -> str:
        [database, schema, compute_pool] = configuration.compute_pool.split(".")

        job_manifest_yaml = yaml.dump(configuration.job_manifest)
        job_service_name = self._slugify_service_name(configuration.name, flow_run.id)

        self._logger.info(
            f"Starting job service {job_service_name} in compute pool {compute_pool}..."
        )

        job_service = JobService(
            name=job_service_name,
            compute_pool=compute_pool,
            spec=ServiceSpec(job_manifest_yaml),
            external_access_integrations=configuration.external_access_integrations
            if configuration.external_access_integrations
            else None,  # Don't set if list is empty.
            query_warehouse=configuration.query_warehouse,
            comment=configuration.service_comment,
        )

        root.databases[database].schemas[schema].services.execute_job_async(job_service)

        return job_service_name

    def _watch_service(
        self,
        job_service_name: str,
        root: Root,
        configuration: SPCSWorkerConfiguration,
    ) -> int:
        pool_start_datetime = datetime.datetime.now(datetime.timezone.utc)
        pool_start_time_seconds = pool_start_datetime.timestamp()
        pool_timeout = configuration.pool_start_timeout_seconds

        [database, schema, compute_pool] = configuration.compute_pool.split(".")
        pool = root.compute_pools[compute_pool]

        while True:
            pool_state = pool.fetch().state

            # Wait until the compute pool is active. It might be idle, resizing, etc.
            if pool_state == "ACTIVE":
                break

            elapsed_time = time.time() - pool_start_time_seconds

            if pool_timeout and elapsed_time > pool_timeout:
                raise RuntimeError(
                    f"Timed out after {elapsed_time} s while waiting for compute pool start."
                )

            self._logger.info(
                f"Compute pool {compute_pool} is in state {pool_state}, checking for ACTIVE state again in {configuration.service_watch_poll_interval} seconds."
            )
            time.sleep(configuration.service_watch_poll_interval)

        service_start_datetime = datetime.datetime.now(datetime.timezone.utc)
        service_start_time_seconds = service_start_datetime.timestamp()
        service_timeout = configuration.service_start_timeout_seconds

        service = root.databases[database].schemas[schema].services[job_service_name]
        last_log_time = service_start_datetime

        while True:
            # Sleep first, give the job service a chance to start.
            time.sleep(configuration.service_watch_poll_interval)

            try:
                # There should only be one container in the service.
                container = next(service.get_containers())
                service_status = container.service_status

                # If status == PENDING or similar, we'll get an exception if we try to retrieve logs.
                if configuration.stream_output and service_status in (
                    "RUNNING",
                    "DONE",
                    "FAILED",
                ):
                    last_log_time = self._get_and_stream_output(
                        service=service,
                        container=container,
                        last_log_time=last_log_time,
                    )

                # If status in one of these, the container is no longer running.
                if service_status in (
                    "DONE",
                    "FAILED",
                    "SUSPENDING",
                    "SUSPENDED",
                    "DELETING",
                    "DELETED",
                    "INTERNAL_ERROR",
                ):
                    break
            except NotFoundError as not_found_error:
                elapsed_time = time.time() - service_start_time_seconds

                if service_timeout and elapsed_time > service_timeout:
                    raise RuntimeError(
                        f"Timed out after {elapsed_time} s while waiting for service start."
                    ) from not_found_error

                self._logger.info(
                    f"Service {job_service_name} isn't running yet, polling for status again in {configuration.service_watch_poll_interval} seconds."
                )

        return 0

    @staticmethod
    def _get_snowflake_connection_parameters(
        configuration: SPCSWorkerConfiguration,
    ) -> dict[str, Any]:
        """Get the Snowflake connection parameters for the worker.

        Args:
            configuration: The worker configuration.

        Returns:
            A dictionary containing the Snowflake connection parameters.

        """
        # If SNOWFLAKE_HOST is defined, we're probably running in Snowflake.
        # That means we can run as the service account.
        if os.getenv("SNOWFLAKE_HOST"):
            connection_parameters = {
                "host": os.getenv("SNOWFLAKE_HOST"),
                "account": os.getenv("SNOWFLAKE_ACCOUNT"),
                "token": Path("/snowflake/session/token").read_text(),
                "authenticator": "oauth",
            }
        else:
            connection_parameters = {
                "account": configuration.snowflake_credentials.account,
                "user": configuration.snowflake_credentials.user,
                "private_key": configuration.snowflake_credentials.resolve_private_key(),
                "role": configuration.snowflake_credentials.role,
            }

        return connection_parameters

    @staticmethod
    def _slugify_service_name(service_name: str, flow_run_id: UUID) -> Optional[str]:
        """Generates a service name to match the configured name, ensuring it's SPCS compatible."""
        # Must match `/?[a-zA-Z0-9][a-zA-Z0-9_]+` in the end.
        if not service_name:
            return None

        # Convert flow_run_id to a string.
        flow_run_id = str(flow_run_id)

        # Slugify flow_name and ensure that the generated name won't be too long for the max deployment name length (63).
        slugified_flow_name = slugify(
            service_name,
            separator="_",
            max_length=62 - len(flow_run_id),
            regex_pattern=r"[^a-zA-Z0-9_]+",  # Only allow lowercase, uppercase letters, numbers, underscores.
        ).strip("_")  # Remove leading and trailing underscores.

        return f"{slugified_flow_name}_{flow_run_id.replace('-', '_')}"

    def _get_and_stream_output(
        self,
        service: ServiceResource,
        container: ServiceContainer,
        last_log_time: datetime,
    ) -> datetime:
        """Fetches logs output from the job container and writes all entries after a given time to stderr.

        Args:
            service: The service we want to retrieve logs for.
            container: The service container.
            last_log_time: The timestamp of the last output line already streamed.

        Returns:
            The time of the most recent output line written by this call.

        """
        logs = self._get_logs(service=service, container=container)

        return self._stream_output(logs, last_log_time)

    def _get_logs(
        self,
        service: ServiceResource,
        container: ServiceContainer,
        max_lines: int = 100,
    ) -> str:
        """Gets the most recent service logs up to a given maximum.

        Args:
            service: The service we want to retrieve logs for.
            container: The service container.
            max_lines: The number of log lines to pull. Defaults to 100.

        Returns:
            A string containing the requested log entries, one per line.

        """
        instance_id = container.instance_id
        container_name = container.container_name

        return service.get_service_logs(
            instance_id=instance_id, container_name=container_name, num_lines=max_lines
        )

    def _stream_output(self, log_content: str, last_log_time: datetime) -> datetime:
        """Writes each entry from a string of log lines to stderr.

        Args:
            log_content: A string containing SPCS container logs.
            last_log_time: The timestamp of the last output line already streamed.

        Returns:
            The time of the most recent output line written by this call.

        """
        if not log_content:
            # Nothing to stream.
            return last_log_time

        log_lines = log_content.split("\n")

        last_written_time = last_log_time

        for log_line in log_lines:
            # Skip if the line is blank or whitespace.
            if not log_line.strip():
                continue

            line_parts = log_line.split(" ")
            # The timestamp should always be before the first space in the line.
            line_timestamp = line_parts[0]

            try:
                # Parse the timestamp and make it timezone-aware.
                line_time = dateutil.parser.parse(line_timestamp).replace(
                    tzinfo=datetime.timezone.utc
                )

                if line_time > last_written_time:
                    print(log_line, file=sys.stderr)
                    last_written_time = line_time
            except dateutil.parser.ParserError as e:
                self._logger.debug(
                    ("Unable to parse timestamp from SPCS log line: %s"),
                    log_line,
                    exc_info=e,
                )

        return last_written_time



import os
from unittest.mock import MagicMock

import pytest
from prefect_snowflake.credentials import SnowflakeCredentials

from prefect.testing.utilities import prefect_test_harness


def _read_test_file(name: str) -> bytes:
    """
    Args:
        name: File to load from test_data folder.

    Returns:
        File content as binary.
    """
    full_name = os.path.join(os.path.split(__file__)[0], "test_data", name)
    with open(full_name, "rb") as fd:
        return fd.read()


@pytest.fixture(scope="session", autouse=True)
def prefect_db():
    """
    Sets up test harness for temporary DB during test runs.
    """
    try:
        with prefect_test_harness():
            yield
    except OSError as e:
        if "Directory not empty" in str(e):
            pass
        else:
            raise e


@pytest.fixture()
def credentials_params() -> dict[str, str]:
    return {
        "account": "account",
        "user": "user",
        "password": "password",
    }


@pytest.fixture()
def private_credentials_params() -> dict[str, str | bytes]:
    return {
        "account": "account",
        "user": "user",
        "password": "letmein",
        "private_key": _read_test_file("test_cert.p8"),
    }


@pytest.fixture()
def private_key_path_credentials_params() -> dict[str, str]:
    return {
        "account": "account",
        "user": "user",
        "private_key_path": "path/to/private/key",
        "private_key_passphrase": "letmein",
    }


@pytest.fixture()
def private_no_pass_credentials_params() -> dict[str, str | bytes]:
    return {
        "account": "account",
        "user": "user",
        "password": "letmein",
        "private_key": _read_test_file("test_cert_no_pass.p8"),
    }


@pytest.fixture()
def private_no_pass_connector_params(private_no_pass_credentials_params):
    snowflake_credentials = SnowflakeCredentials(**private_no_pass_credentials_params)
    _connector_params = {
        "schema": "schema_input",
        "database": "database",
        "warehouse": "warehouse",
        "credentials": snowflake_credentials,
    }
    return _connector_params


@pytest.fixture()
def private_malformed_credentials_params() -> dict[str, str | bytes]:
    return {
        "account": "account",
        "user": "user",
        "password": "letmein",
        "private_key": _read_test_file("test_cert_malformed_format.p8"),
    }


@pytest.fixture()
def workload_identity_credentials_params() -> dict[str, str]:
    return {
        "account": "account",
        "authenticator": "workload_identity",
        "workload_identity_provider": "aws",
    }


@pytest.fixture()
def workload_identity_oidc_credentials_params() -> dict[str, str]:
    return {
        "account": "account",
        "authenticator": "workload_identity",
        "workload_identity_provider": "oidc",
        "token": "token",
    }


@pytest.fixture(autouse=True)
def snowflake_connect_mock(monkeypatch):
    mock_cursor = MagicMock(name="cursor mock")
    results = iter([0, 1, 2, 3, 4])
    mock_cursor.return_value.fetchone.side_effect = lambda: (next(results),)
    mock_cursor.return_value.fetchmany.side_effect = lambda size: list(
        (next(results),) for i in range(size)
    )
    mock_cursor.return_value.fetchall.side_effect = lambda: [
        (result,) for result in results
    ]

    mock_connection = MagicMock(name="connection mock")

    def close_connection(*args, **kwargs):
        """Mock function to simulate closing the connection."""
        mock_connection.return_value.is_closed.return_value = True

    mock_connection.return_value.is_closed.return_value = False
    mock_connection.return_value.__exit__.side_effect = close_connection
    mock_connection.return_value.is_still_running.return_value = False
    mock_connection.return_value.cursor = mock_cursor

    monkeypatch.setattr("snowflake.connector.connect", mock_connection)
    return mock_connection



import uuid
from unittest.mock import AsyncMock, Mock

import pytest
from prefect_snowflake import SnowflakeCredentials
from prefect_snowflake.experimental.workers.spcs import (
    SPCSServiceTemplateVariables,
    SPCSWorker,
    SPCSWorkerConfiguration,
)
from pydantic import SecretStr

from prefect.client.schemas import FlowRun
from prefect.server.schemas.core import Flow
from prefect.utilities.dockerutils import get_prefect_image_name


# Helper functions
async def create_job_configuration(
    snowflake_credentials, worker_flow_run, overrides=None, run_prep=True
):
    """
    Returns a basic initialized SPCS infrastructure block suitable for use in a variety of tests.
    """
    if overrides is None:
        overrides = {}

    values = {
        "command": "test",
        "env": {},
        "snowflake_credentials": snowflake_credentials,
        "compute_pool": "common.compute.test_pool",
        "name": None,
        "service_watch_poll_interval": 1,
        "stream_output": False,
    }

    for k, v in overrides.items():
        values = {**values, k: v}

    job_service_variables = SPCSServiceTemplateVariables(**values)

    json_config = {
        "job_configuration": SPCSWorkerConfiguration.json_template(),
        "variables": job_service_variables.model_dump(),
    }

    job_service_configuration = await SPCSWorkerConfiguration.from_template_and_values(
        json_config, values
    )

    if run_prep:
        job_service_configuration.prepare_for_flow_run(worker_flow_run)

    return job_service_configuration


# Fixtures
@pytest.fixture
def snowflake_credentials():
    account = "test_account"
    user = "test_user"
    password = "test_password"

    return SnowflakeCredentials(account=account, user=user, password=password)


@pytest.fixture
def mock_prefect_client(worker_flow):
    """
    A fixture that provides a mock Prefect client
    """
    mock_client = Mock()
    mock_client.read_flow = AsyncMock()
    mock_client.read_flow.return_value = worker_flow

    return mock_client


@pytest.fixture
def worker_flow():
    return Flow(id=uuid.uuid4(), name="test-flow")


@pytest.fixture
def worker_flow_run(worker_flow):
    return FlowRun(id=uuid.uuid4(), flow_id=worker_flow.id, name="test-flow-run")


# Tests
async def test_worker_valid_command_validation(snowflake_credentials, worker_flow_run):
    # ensure the validator allows valid commands to pass through
    command = "command arg1 arg2"

    spcs_job_config = await create_job_configuration(
        snowflake_credentials, worker_flow_run, {"command": command}
    )

    assert spcs_job_config.command == command


def test_worker_invalid_command_validation(snowflake_credentials):
    # ensure invalid commands cause a validation error
    with pytest.raises(ValueError):
        SPCSWorkerConfiguration(
            command=["invalid_command", "arg1", "arg2"],
            subscription_id=SecretStr("test"),
            resource_group_name="test",
            snowflake_credentials=snowflake_credentials,
        )


async def test_job_configuration_creation(snowflake_credentials, worker_flow_run):
    config = {
        "snowflake_credentials": snowflake_credentials,
        "image": "my-image",
        "image_registry": "my_image_registry",
        "compute_pool": "common.compute.test_pool",
        "external_access_integrations": ["test_integration"],
        "cpu_request": "1",
        "memory_request": "1G",
        "env": {"TEST": "VALUE"},
    }

    spcs_job_config = await create_job_configuration(
        snowflake_credentials, worker_flow_run, config
    )

    assert spcs_job_config.image == "my-image"
    assert spcs_job_config.image_registry == "my_image_registry"
    assert spcs_job_config.cpu_request == "1"
    assert spcs_job_config.memory_request == "1G"
    assert spcs_job_config.env.get("TEST") == "VALUE"
    assert spcs_job_config.compute_pool == "common.compute.test_pool"
    assert spcs_job_config.external_access_integrations == ["test_integration"]
    assert (
        spcs_job_config.snowflake_credentials.account == snowflake_credentials.account
    )
    assert spcs_job_config.snowflake_credentials.user == snowflake_credentials.user
    assert (
        spcs_job_config.snowflake_credentials.password == snowflake_credentials.password
    )


async def test_image_populated_in_template_when_not_provided(
    worker_flow_run, snowflake_credentials
):
    config = await SPCSWorkerConfiguration.from_template_and_values(
        base_job_template=SPCSWorker.get_default_base_job_template(),
        values=SPCSServiceTemplateVariables(
            snowflake_credentials=snowflake_credentials,
            compute_pool="common.compute.test_pool",
        ).model_dump(exclude_unset=True),
    )
    config.prepare_for_flow_run(worker_flow_run)

    assert config.image == get_prefect_image_name()
    assert (
        config.job_manifest["spec"]["containers"][0]["image"]
        == get_prefect_image_name()
    )


@pytest.mark.parametrize(
    "flow_name",
    [
        "Short Flow",
        "Another Notification",
        "Extremely Long Flow Name That Exceeds The Maximum Length Allowed",
    ],
)
async def test_consistent_service_job_naming(
    mock_prefect_client: mock_prefect_client,
    worker_flow_run,
    flow_name,
):
    max_length = 63

    flow = Flow(id=worker_flow_run.flow_id, name=flow_name)
    mock_prefect_client.read_flow.return_value = flow

    flow_run = FlowRun(flow_id=flow.id, name=f"{flow_name} run")

    service_job_name = SPCSWorker._slugify_service_name(
        service_name=flow_name, flow_run_id=flow_run.id
    )

    assert " " not in service_job_name
    assert all(c.isalnum() or c in ("-", "_") for c in service_job_name)
    assert service_job_name.endswith(str(flow_run.id).replace("-", "_"))

    name_without_id = service_job_name[:-37]

    assert name_without_id.replace("_", " ").lower() in flow_name.lower()

    assert len(service_job_name) <= max_length, (
        f"Length: {len(service_job_name)}, Max: {max_length}"
    )


async def test_timeout_configuration_defaults(snowflake_credentials, worker_flow_run):
    """Test that timeout configurations have expected default values."""
    spcs_job_config = await create_job_configuration(
        snowflake_credentials, worker_flow_run
    )

    # Test default timeout values
    assert spcs_job_config.pool_start_timeout_seconds == 600
    assert spcs_job_config.service_start_timeout_seconds == 300


async def test_timeout_configuration_custom_values(
    snowflake_credentials, worker_flow_run
):
    """Test that custom timeout values are properly set."""
    config_overrides = {
        "pool_start_timeout_seconds": 120,
        "service_start_timeout_seconds": 60,
    }

    spcs_job_config = await create_job_configuration(
        snowflake_credentials, worker_flow_run, config_overrides
    )

    assert spcs_job_config.pool_start_timeout_seconds == 120
    assert spcs_job_config.service_start_timeout_seconds == 60


@pytest.mark.parametrize(
    "timeout_seconds,expected_valid",
    [
        (0, True),  # Zero timeout means no timeout
        (1, True),  # Minimum positive timeout
        (3600, True),  # One hour timeout
        (-1, False),  # Negative timeouts should be invalid
    ],
)
async def test_timeout_validation(
    snowflake_credentials, worker_flow_run, timeout_seconds, expected_valid
):
    """Test validation of timeout values."""
    config_overrides = {
        "pool_start_timeout_seconds": timeout_seconds,
        "service_start_timeout_seconds": timeout_seconds,
    }

    if expected_valid:
        # Should not raise an exception
        config = await create_job_configuration(
            snowflake_credentials, worker_flow_run, config_overrides
        )
        assert config.pool_start_timeout_seconds == timeout_seconds
        assert config.service_start_timeout_seconds == timeout_seconds
    else:
        # Should raise a validation error when prepare_for_flow_run is called
        with pytest.raises((ValueError, Exception)):  # Catch validation errors
            await create_job_configuration(
                snowflake_credentials, worker_flow_run, config_overrides
            )



import inspect
from unittest.mock import MagicMock

import pytest
from prefect_snowflake.credentials import SnowflakeCredentials
from prefect_snowflake.database import (
    BEGIN_TRANSACTION_STATEMENT,
    END_TRANSACTION_STATEMENT,
    SnowflakeConnector,
    snowflake_multiquery,
    snowflake_multiquery_async,
    snowflake_query,
    snowflake_query_async,
)
from pydantic import SecretBytes, SecretStr
from snowflake.connector import DictCursor
from snowflake.connector.cursor import SnowflakeCursor as OriginalSnowflakeCursorClass

from prefect import flow


@pytest.fixture
def connector_params(credentials_params):
    snowflake_credentials = SnowflakeCredentials(**credentials_params)
    _connector_params = {
        "schema": "schema_input",
        "database": "database",
        "warehouse": "warehouse",
        "credentials": snowflake_credentials.model_dump(),
    }
    return _connector_params


@pytest.fixture
def private_connector_params(private_credentials_params):
    snowflake_credentials = SnowflakeCredentials(**private_credentials_params)
    _connector_params = {
        "schema": "schema_input",
        "database": "database",
        "warehouse": "warehouse",
        "credentials": snowflake_credentials.model_dump(),
    }
    return _connector_params


async def run_sync_or_async(func, *args, **kwargs):
    result = func(*args, **kwargs)
    if inspect.isawaitable(result):
        return await result
    return result


def test_snowflake_connector_init(connector_params):
    snowflake_connector = SnowflakeConnector(**connector_params)
    actual_connector_params = snowflake_connector.model_dump()
    for param in connector_params:
        expected = connector_params[param]
        if param == "schema":
            param = "schema_"
        actual = actual_connector_params[param]
        if isinstance(actual, (SecretStr, SecretBytes)):
            actual = actual.get_secret_value()
        assert actual == expected


def test_snowflake_connector_password_is_secret_str(connector_params):
    snowflake_connector = SnowflakeConnector(**connector_params)
    password = snowflake_connector.credentials.password
    assert isinstance(password, SecretStr)
    assert password.get_secret_value() == "password"


def test_snowflake_connector_private_key_is_secret(private_connector_params):
    snowflake_connector = SnowflakeConnector(**private_connector_params)
    private_key = snowflake_connector.credentials.private_key
    assert isinstance(private_key, (SecretStr, SecretBytes))


class SnowflakeCursor:
    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def execute_async(self, query, params):
        query_id = "1234"
        self.result = {query_id: [(query, params)]}
        return {"queryId": query_id}

    def get_results_from_sfqid(self, query_id):
        self.query_result = self.result[query_id]

    def fetchall(self):
        return self.query_result

    def execute(self, query, params=None):
        self.query_result = [(query, params, "sync")]
        return self


class SnowflakeConnection:
    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def cursor(self, cursor_type):
        return SnowflakeCursor()

    def is_still_running(self, state):
        return state

    def get_query_status_throw_if_error(self, query_id):
        return False


@pytest.fixture()
def snowflake_connector(snowflake_connect_mock):
    snowflake_connector_mock = MagicMock()
    snowflake_connector_mock.get_connection.return_value = SnowflakeConnection()
    return snowflake_connector_mock


@pytest.mark.parametrize("query_function", [snowflake_query, snowflake_query_async])
async def test_snowflake_query(query_function, snowflake_connector):
    @flow
    async def test_flow():
        result = await run_sync_or_async(
            query_function, "query", snowflake_connector, params=("param",)
        )
        return result

    result = await run_sync_or_async(test_flow)
    assert result[0][0] == "query"
    assert result[0][1] == ("param",)


@pytest.mark.parametrize(
    "query_function", [snowflake_multiquery, snowflake_multiquery_async]
)
async def test_snowflake_multiquery(query_function, snowflake_connector):
    @flow
    async def test_flow():
        result = await run_sync_or_async(
            query_function, ["query1", "query2"], snowflake_connector, params=("param",)
        )
        return result

    result = await run_sync_or_async(test_flow)
    assert result[0][0][0] == "query1"
    assert result[0][0][1] == ("param",)
    assert result[1][0][0] == "query2"
    assert result[1][0][1] == ("param",)


@pytest.mark.parametrize(
    "query_function", [snowflake_multiquery, snowflake_multiquery_async]
)
async def test_snowflake_multiquery_transaction(query_function, snowflake_connector):
    @flow
    async def test_flow():
        result = await run_sync_or_async(
            query_function,
            ["query1", "query2"],
            snowflake_connector,
            params=("param",),
            as_transaction=True,
        )
        return result

    result = await run_sync_or_async(test_flow)
    assert result[0][0][0] == "query1"
    assert result[0][0][1] == ("param",)
    assert result[1][0][0] == "query2"
    assert result[1][0][1] == ("param",)


@pytest.mark.parametrize(
    "query_function", [snowflake_multiquery, snowflake_multiquery_async]
)
async def test_snowflake_multiquery_transaction_with_transaction_control_results(
    query_function, snowflake_connector
):
    @flow
    async def test_flow():
        result = await run_sync_or_async(
            query_function,
            ["query1", "query2"],
            snowflake_connector,
            params=("param",),
            as_transaction=True,
            return_transaction_control_results=True,
        )
        return result

    result = await run_sync_or_async(test_flow)
    assert result[0][0][0] == BEGIN_TRANSACTION_STATEMENT
    assert result[1][0][0] == "query1"
    assert result[1][0][1] == ("param",)
    assert result[2][0][0] == "query2"
    assert result[2][0][1] == ("param",)
    assert result[3][0][0] == END_TRANSACTION_STATEMENT


def test_snowflake_private_connector_init(private_connector_params):
    snowflake_connector = SnowflakeConnector(**private_connector_params)
    actual_connector_params = snowflake_connector.model_dump()
    for param in private_connector_params:
        expected = private_connector_params[param]
        if param == "schema":
            param = "schema_"
        actual = actual_connector_params[param]
        if isinstance(actual, (SecretStr, SecretBytes)):
            actual = actual.get_secret_value()
        assert actual == expected


class TestSnowflakeConnector:
    @pytest.fixture
    def snowflake_connector(self, connector_params, snowflake_connect_mock):
        connector = SnowflakeConnector(**connector_params)
        return connector

    def test_block_initialization_handles_schema_alias(self, connector_params):
        # First, test the case where the schema is set with the alias
        connector_params["schema"] = "schema_input"
        connector_params.pop("schema_", None)
        connector = SnowflakeConnector(**connector_params)
        assert connector.schema_ == "schema_input"

        # Now, test the case where the schema is set with the field name
        connector_params["schema_"] = "schema_input"
        connector_params.pop("schema", None)
        connector = SnowflakeConnector(**connector_params)
        assert connector.schema_ == "schema_input"

    def test_block_initialization(self, snowflake_connector):
        assert snowflake_connector._connection is None
        assert snowflake_connector._unique_cursors is None

    def test_get_connection(self, snowflake_connector: SnowflakeConnector, caplog):
        connection = snowflake_connector.get_connection()
        assert snowflake_connector._connection is connection
        assert caplog.records[0].msg == "Started a new connection to Snowflake."

    def test_get_connection_multiple_calls(
        self, snowflake_connector: SnowflakeConnector, caplog
    ):
        connection1 = snowflake_connector.get_connection()
        connection2 = snowflake_connector.get_connection()

        assert connection1 is connection2
        assert caplog.records[0].msg == "Started a new connection to Snowflake."
        assert len(caplog.records) == 1

    def test_get_connection_context_manager_multiple_calls(
        self, snowflake_connector: SnowflakeConnector, caplog
    ):
        with snowflake_connector.get_connection():
            pass

        with snowflake_connector.get_connection():
            pass

        assert len(caplog.records) == 2
        assert caplog.records[0].msg == "Started a new connection to Snowflake."
        assert caplog.records[1].msg == "Started a new connection to Snowflake."

    def test_reset_cursors(self, snowflake_connector: SnowflakeConnector, caplog):
        mock_cursor = MagicMock()
        snowflake_connector.reset_cursors()
        assert caplog.records[0].msg == "There were no cursors to reset."

        snowflake_connector._start_connection()
        snowflake_connector._unique_cursors["12345"] = mock_cursor
        snowflake_connector.reset_cursors()
        assert len(snowflake_connector._unique_cursors) == 0
        mock_cursor.close.assert_called_once()

    @pytest.mark.parametrize("fetch_function_name", ["fetch_one", "fetch_one_async"])
    async def test_fetch_one(self, fetch_function_name, snowflake_connector):
        fetch_function = getattr(snowflake_connector, fetch_function_name)
        result = await run_sync_or_async(fetch_function, "SELECT 1", parameters=None)
        assert result == (0,)

    def test_fetch_one_cursor_set_to_dict_cursor(
        self, snowflake_connector: SnowflakeConnector
    ):
        _ = snowflake_connector.fetch_one(
            "query", parameters=("param",), cursor_type=DictCursor
        )
        args, _ = snowflake_connector._connection.cursor.call_args
        assert args[0] == DictCursor

    def test_fetch_one_cursor_default(self, snowflake_connector: SnowflakeConnector):
        _ = snowflake_connector.fetch_one("query", parameters=("param",))
        args, kwargs = snowflake_connector._connection.cursor.call_args

        assert args[0] == OriginalSnowflakeCursorClass

    def test_fetch_all_cursor_set_to_dict_cursor(
        self, snowflake_connector: SnowflakeConnector
    ):
        _ = snowflake_connector.fetch_all(
            "query", parameters=("param",), cursor_type=DictCursor
        )
        args, _ = snowflake_connector._connection.cursor.call_args

        assert args[0] == DictCursor

    def test_fetch_all_cursor_default(self, snowflake_connector: SnowflakeConnector):
        _ = snowflake_connector.fetch_all("query", parameters=("param",))
        args, _ = snowflake_connector._connection.cursor.call_args

        assert args[0] == OriginalSnowflakeCursorClass

    def test_fetch_all_executes_directly_on_cursor(
        self, snowflake_connector: SnowflakeConnector
    ):
        """Test that fetch_all executes directly on the cursor instead of using self.execute."""
        snowflake_connector._start_connection()
        mock_cursor = MagicMock()
        snowflake_connector._connection.cursor.return_value = mock_cursor
        mock_cursor.fetchall.return_value = [(1,)]

        result = snowflake_connector.fetch_all("SELECT 1")

        mock_cursor.execute.assert_called_once_with("SELECT 1", params=None)
        mock_cursor.fetchall.assert_called_once()
        assert result == [(1,)]

    def test_fetch_one_executes_directly_on_cursor(
        self, snowflake_connector: SnowflakeConnector
    ):
        """Test that fetch_one executes directly on the cursor instead of using self.execute."""
        snowflake_connector._start_connection()
        mock_cursor = MagicMock()
        snowflake_connector._connection.cursor.return_value = mock_cursor
        mock_cursor.fetchone.return_value = (1,)

        result = snowflake_connector.fetch_one("SELECT 1")

        mock_cursor.execute.assert_called_once_with("SELECT 1", params=None)
        mock_cursor.fetchone.assert_called_once()
        assert result == (1,)

    def test_fetch_many_executes_directly_on_cursor(
        self, snowflake_connector: SnowflakeConnector
    ):
        """Test that fetch_many executes directly on the cursor instead of using self.execute."""
        snowflake_connector._start_connection()
        mock_cursor = MagicMock()
        snowflake_connector._connection.cursor.return_value = mock_cursor
        mock_cursor.fetchmany.return_value = [(1,), (2,)]

        result = snowflake_connector.fetch_many("SELECT 1", size=2)

        mock_cursor.execute.assert_called_once_with("SELECT 1", params=None)
        mock_cursor.fetchmany.assert_called_once_with(size=2)
        assert result == [(1,), (2,)]

    @pytest.mark.parametrize("fetch_function_name", ["fetch_many", "fetch_many_async"])
    async def test_fetch_many(self, fetch_function_name, snowflake_connector):
        fetch_function = getattr(snowflake_connector, fetch_function_name)
        result = await run_sync_or_async(
            fetch_function, "SELECT 1", parameters=None, size=3
        )
        assert result == [(0,), (1,), (2,)]

    @pytest.mark.parametrize("fetch_function_name", ["fetch_all", "fetch_all_async"])
    async def test_fetch_all(self, fetch_function_name, snowflake_connector):
        fetch_function = getattr(snowflake_connector, fetch_function_name)
        result = await run_sync_or_async(fetch_function, "SELECT 1", parameters=None)
        assert result == [(0,), (1,), (2,), (3,), (4,)]

    def test_execute(self, snowflake_connector: SnowflakeConnector):
        assert snowflake_connector.execute("query", parameters=("param",)) is None

    @pytest.mark.parametrize(
        "execute_function_name", ["execute_many", "execute_many_async"]
    )
    async def test_execute_many(self, execute_function_name, snowflake_connector):
        execute_function = getattr(snowflake_connector, execute_function_name)
        await run_sync_or_async(
            execute_function,
            "INSERT INTO table VALUES (%(val)s)",
            seq_of_parameters=[{"val": 1}, {"val": 2}],
        )

    def test_close(self, snowflake_connector: SnowflakeConnector, caplog):
        assert snowflake_connector.close() is None
        assert caplog.records[0].msg == "There were no cursors to reset."
        assert caplog.records[1].msg == "There was no connection open to be closed."

        snowflake_connector._start_connection()
        assert snowflake_connector.close() is None
        assert snowflake_connector._connection is None
        assert snowflake_connector._unique_cursors == {}

    def test_context_management(self, snowflake_connector):
        with snowflake_connector:
            assert snowflake_connector._connection is None
            assert snowflake_connector._unique_cursors is None

        assert snowflake_connector._connection is None
        assert snowflake_connector._unique_cursors is None

    async def test_fetch_one_async(self, snowflake_connector):
        async def fetch_one_async_wrapper():
            return await snowflake_connector.fetch_one_async(
                "SELECT 1", parameters=None
            )

        result = await fetch_one_async_wrapper()
        assert result == (0,)

    async def test_fetch_many_async(self, snowflake_connector):
        async def fetch_many_async_wrapper():
            return await snowflake_connector.fetch_many_async(
                "SELECT 1", parameters=None, size=3
            )

        result = await fetch_many_async_wrapper()
        assert result == [(0,), (1,), (2,)]

    async def test_fetch_all_async(self, snowflake_connector):
        async def fetch_all_async_wrapper():
            return await snowflake_connector.fetch_all_async(
                "SELECT 1", parameters=None
            )

        result = await fetch_all_async_wrapper()
        assert result == [(0,), (1,), (2,), (3,), (4,)]

    async def test_execute_many_async(self, snowflake_connector):
        async def execute_many_async_wrapper():
            await snowflake_connector.execute_many_async(
                "INSERT INTO table VALUES (%(val)s)",
                seq_of_parameters=[{"val": 1}, {"val": 2}],
            )

        await execute_many_async_wrapper()
